__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "GPL v3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

from chembridger.dataset.Splitting import Splitting


def loader(
        df,
        num_splits=5,
        num_repeats=1,
        verbose=True,
):
    return Splitting(
        df=df,
        num_repeats=num_repeats,
        num_splits=num_splits,
        verbose=verbose,
    )


def calc_cluster(
        loader,
        cluster_met='random_cluster',
):
    return loader.calc_cluster_sgl(
        cluster_met=cluster_met,
    )


def calc_clusters(
        loader,
):
    return loader.calc_cluster_all()


def calc_train_test_ids(
        loader,
):
    return loader.train_test_sample_ids()


def generate_summary(
        loader,
        train_test_ids,
):
    return loader.add_summary(
        split_id_dict=train_test_ids,
    )


def generate_mark(
        loader,
        train_test_ids,
):
    return loader.add_marks(
        split_id_dict=train_test_ids,
    )


def cluster_num_selection(
        loader,
        cluster_met="umap_cluster",
        lower_bound=5,
        upper_bound=76,
        step=5,
):
    return loader.cluster_num_sel(
        cluster_met=cluster_met,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
        step=step,
    )


def sim_test_train(
        loader,
        train_test_ids,
        met='tanimoto',
        top_n=5,
):
    return loader.sim_test_train(
        split_id_dict=train_test_ids,
        met=met,
        top_n=top_n,
    )




if __name__ == "__main__":
    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from chembridger.path import to
    import bioaxis as ba

    df = pd.read_csv(to("data/protac/finalPROTAC.csv"))

    # df = df.sample(frac=0.8, replace=True).reset_index(drop=True)
    # print(df)

    loader_p = loader(
        df=df,
        num_repeats=1,
        num_splits=5,
        verbose=False,
    )

    split_id_dict = calc_train_test_ids(
        loader=loader_p,
    )
    # print(split_id_dict)

    df_split_summary = generate_summary(
        loader=loader_p,
        train_test_ids=split_id_dict,
    )
    # print(df_split_summary)

    ### /*** plot distribution size ***/
    # _, ax = plt.subplots(1, 1, figsize=(8, 4), sharey=True)
    # ba.ds.plot_distrib_size(
    #     df=df_split_summary,
    #     ax=ax,
    #     # num_rows=4,
    #     # num_cols=5,
    # )
    # plt.show()

    df_split_mark = generate_mark(
        loader=loader_p,
        train_test_ids=split_id_dict,
    )
    # print(df_split_mark)

    df_cluster = calc_cluster(
        loader=loader_p,
    )
    # print(df_cluster)

    df_clusters = calc_clusters(
        loader=loader_p,
    )
    # print(df_clusters)

    ### /*** mols ***/
    from smiles import to_mols
    df['mol'] = to_mols(
        smiles=df["SMILES"],
        verbose=True,
    )

    ### /*** fps ***/
    from desc import fingerprint
    df['fp'] = fingerprint(
        mols=df["mol"].values.tolist(),
        met="Morgan",
    )

    # df_clus_sel = cluster_num_selection(
    #     loader=loader_p,
    #     cluster_met="umap_cluster",
    #     lower_bound=5,
    #     upper_bound=76,
    #     step=5,
    # )
    # print(df_clus_sel)

    ### /*** plot num_clusters ***/
    # _, ax = plt.subplots(1, 1, figsize=(10, 5), sharey=True)
    # ba.ds.plot_num_clusters(
    #     df=df_clus_sel,
    #     ax=ax,
    #     # num_rows=4,
    #     # num_cols=5,
    # )
    # plt.show()


    x_coord, y_coord = ba.dr.get_2d_coords(
        feature=np.stack(df['fp']),
        met="tsne",
    )
    df['x'] = x_coord
    df['y'] = y_coord
    # print(df)

    df = pd.concat([df_split_mark, df], axis=1).reset_index(drop=True)
    # print(df)

    ### /*** plot dim reduction ***/
    _, ax = plt.subplots(4, 5, figsize=(15, 12), sharey=True)
    ba.ds.plot_dim_reduction(
        df=df,
        ax=ax,
        # num_rows=4,
        # num_cols=5,
        cluster_mets=loader_p.cluster_mets,
        num_splits=5,
    )
    plt.show()

    ### /*** plot mol similairty ***/
    # df_sim = sim_test_train(
    #     loader=loader_p,
    #     train_test_ids=split_id_dict,
    #     met='tanimoto',
    # )
    # print(df_sim)
    #
    # _, ax = plt.subplots(1, 1, figsize=(10, 5), sharey=True)
    # ba.ds.plot_mol_similairty(
    #     df=df_sim,
    #     ax=ax,
    # )
    # plt.show()


__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2022"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"

import os


def root_dict():
    """
    ..  @description:
        -------------
        abs file path.

    :return:
    """
    ROOT_DICT = os.path.dirname(os.path.abspath(__file__))
    return ROOT_DICT


def to(path):
    return os.path.join(
        root_dict(),
        path
    )


__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"

import pandas as pd
from rdkit import Chem
from chembridger.util.Console import Console


class SMILES:

    def __init__(
            self,
            smiles,
            verbose=True,
            **kwargs,
    ):
        self.smiles = smiles
        self.df = pd.DataFrame(self.smiles, columns=["SMILES"])

        self.console = Console()
        self.console.verbose = verbose

    @property
    def mols(self, ):
        return self.df.SMILES.apply(Chem.MolFromSmiles)




__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"

from rdkit import Chem
from chembridger.util.Console import Console


class SDF:

    def __init__(
            self,
            sdf_fpn,
            kind='all',
            verbose=False,
    ):
        self.console = Console()
        self.verbose = verbose
        self.console.verbose = self.verbose
        self._kind = kind
        self.sdf_fpn = sdf_fpn
        self.all_mols = self.__all()
        self.console.print('===>Fetching mol info...')

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, value):
        self.console.print('======>Please note that this is from an external attempt')
        if value not in ['all', 'approved', 'expt', 'all_2d', 'approved_2d', 'expt_2d', 'all_dict', 'approved_dict',
                         'expt_dict']:
            raise ValueError(
                'kind has not yet reached there.',
                '| all: ',
                '| approved: ',
                '| expt: ',
                '| all_2d: ',
                '| approved_2d: ',
                '| expt_2d: ',
                '| all_dict: ',
                '| approved_dict: ',
                '| expt_dict: ',
            )
        else:
            self._kind = value

    def __all(self, ):
        return Chem.SDMolSupplier(self.sdf_fpn)

    def extract(self, ):
        if self._kind == 'all':
            self.console.print('=========>You are attempting to get all molecules (all).')
            return [mol for mol in self.all_mols if mol]
        elif self._kind == 'approved':
            self.console.print('=========>You are attempting to get approved molecules (approved).')
            return [mol for mol in self.all_mols if mol and 'approved' in mol.GetProp('DRUG_GROUPS')]
        elif self._kind == 'expt':
            self.console.print('=========>You are attempting to get approved molecules (expt).')
            return [mol for mol in self.all_mols if mol and 'experimental' in mol.GetProp('DRUG_GROUPS')]
        elif self._kind == 'all_2d':
            self.console.print('=========>You are attempting to get all name<->molecules (all_2d).')
            return [[str(mol.GetProp('DATABASE_ID')), mol] for mol in self.all_mols if mol]
        elif self._kind == 'approved_2d':
            self.console.print('=========>You are attempting to get approved name<->molecules (approved_2d).')
            return [[str(mol.GetProp('DATABASE_ID')), mol] for mol in self.all_mols if
                    mol and 'approved' in mol.GetProp('DRUG_GROUPS')]
        elif self._kind == 'expt_2d':
            self.console.print('=========>You are attempting to get experimental name<->molecules (expt_2d).')
            return [[str(mol.GetProp('DATABASE_ID')), mol] for mol in self.all_mols if
                    mol and 'experimental' in mol.GetProp('DRUG_GROUPS')]
        elif self._kind == 'all_dict':
            self.console.print('=========>You are attempting to get all molecules dictionary (all_dict).')
            mols = dict()
            for mol in self.all_mols:
                if mol:
                    # self.console.print(mol.GetProp('DRUG_GROUPS'))
                    mols[mol.GetProp('DATABASE_ID')] = mol
            return mols
        elif self._kind == 'approved_dict':
            self.console.print('=========>You are attempting to get approved molecules dictionary (approved_dict).')
            mols = dict()
            for mol in self.all_mols:
                if mol and 'approved' in mol.GetProp('DRUG_GROUPS'):
                    # self.console.print(mol.GetProp('DRUG_GROUPS'))
                    mols[mol.GetProp('DATABASE_ID')] = mol
            return mols
        elif self._kind == 'expt_dict':
            self.console.print('=========>You are attempting to get approved molecules dictionary (approved_dict).')
            mols = dict()
            for mol in self.all_mols:
                if mol and 'experimental' in mol.GetProp('DRUG_GROUPS'):
                    # print(mol.GetProp('DRUG_GROUPS'))
                    mols[mol.GetProp('DATABASE_ID')] = mol
            return mols


if __name__ == "__main__":
    from chembridger.path import to

    p = SDF(sdf_fpn=to('data/chem/approved.sdf'))
    # p.kind = 'expt_dict' # expt approved all
    p.kind = 'approved'
    print(p.kind)
    # print(p.extract())
    print(len(p.extract()))


__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "GPL v3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

from chembridger.molecule.SMILES import SMILES


def to_mols(
        smiles,
        verbose=True,
):
    return SMILES(
        smiles=smiles,
        verbose=verbose,
    ).mols


if __name__ == "__main__":
    from chembridger.path import to
    import pandas as pd

    df = pd.read_csv(to("data/protac/finalPROTAC.csv"))
    # loader_p = loader(
    #     smiles=df["SMILES"],
    #     verbose=False,
    # )
    df['mol'] = to_mols(
        smiles=df["SMILES"],
        verbose=False,
    )
    print(df['mol'])



__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__email__ = "jianfeng.sunmt@gmail.com"

import numpy as np
import pandas as pd

from tqdm.auto import tqdm

from rdkit.DataStructs import BulkTanimotoSimilarity
import useful_rdkit_utils as uru

from chembridger.util.Console import Console


class Splitting:

    def __init__(
            self,
            df,
            num_splits=5,
            num_repeats=1,
            verbose=True,
    ):
        """

        See Also
        --------
        Original author: Pat Walters
        Post: https://practicalcheminformatics.blogspot.com/2024/11/some-thoughts-on-splitting-chemical.html
        Raw code: https://github.com/PatWalters/practical_cheminformatics_posts/blob/main/splitting/dataset_splitting.ipynb

        Parameters
        ----------
        df
        num_repeats
        """
        self.df = df
        self.num_repeats = num_repeats
        self.num_splits = num_splits

        self.cluster_dict = {
            "random_cluster": uru.get_random_clusters,
            "butina_cluster": uru.get_butina_clusters,
            "umap_cluster": uru.get_umap_clusters,
            "scaffold_cluster": uru.get_bemis_murcko_clusters,
        }
        self.cluster_mets = [*self.cluster_dict.keys()]

        self.console = Console()
        self.console.verbose = verbose

        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FutureWarning)

    def calc_cluster_sgl(self, cluster_met):
        self.df[cluster_met] = self.cluster_dict[cluster_met](self.df.SMILES)
        return self.df

    def calc_cluster_all(self, ):
        for cluster_met in self.cluster_mets:
            self.df[cluster_met] = self.cluster_dict[cluster_met](self.df.SMILES)
        return self.df

    def train_test_sample_ids(self, ):
        split_id_dict = {}
        for cluster_met in self.cluster_mets:

            self.console.print("===>cluster method: {}".format(cluster_met))

            split_id_dict[cluster_met] = {}
            for repeat_num in tqdm(range(self.num_repeats), desc=cluster_met):

                self.console.print("======>repeat time: {}".format(repeat_num))

                split_id_dict[cluster_met][repeat_num] = {}
                cluster_ids = self.cluster_dict[cluster_met](self.df.SMILES)
                gkfs = uru.GroupKFoldShuffle(n_splits=self.num_splits, shuffle=True)
                for i, (train_idx, test_idx) in enumerate(gkfs.split(X=self.df, groups=cluster_ids)):

                    self.console.print("=========>fold: {}".format(i))

                    split_id_dict[cluster_met][repeat_num][i] = [train_idx.tolist(), test_idx.tolist()]
        # print(split_id_dict)
        return split_id_dict

    def add_summary(
            self,
            split_id_dict,
    ):
        split_arr = []
        for cluster_met in self.cluster_mets:
            for repeat_num in range(self.num_repeats):
                for fold_i in range(self.num_splits):
                    train_size = len(split_id_dict[cluster_met][repeat_num][fold_i][0])
                    test_size = len(split_id_dict[cluster_met][repeat_num][fold_i][1])
                    split_arr.append([cluster_met, repeat_num, fold_i, train_size, test_size])
        return pd.DataFrame(
            split_arr,
            columns=[
                "cluster_met",
                "repeat_num",
                "fold_i",
                "train_size",
                "test_size",
            ]
        )

    def add_marks(
            self,
            split_id_dict,
            fold=0,
    ):
        df = pd.DataFrame()
        for cluster_met in self.cluster_mets:
            for repeat_num in range(self.num_repeats):
                for fold_i in range(self.num_splits):
                    attachment = cluster_met + '_r' + str(repeat_num) + '_s' + str(fold_i)
                    df[attachment] = ["train"] * self.df.shape[0]

                    self.console.print(split_id_dict[cluster_met][repeat_num][fold_i])

                    test_idx = split_id_dict[cluster_met][repeat_num][fold_i][1]

                    self.console.print(test_idx)

                    for t in test_idx:
                        df[attachment].at[t] = "test"
        return df

    def cluster_num_sel(
            self,
            cluster_met="umap_cluster",
            lower_bound=5,
            upper_bound=76,
            step=5,
    ):
        res = []
        for num_clus in tqdm(range(lower_bound, upper_bound, step)):
            for num_repeat in range(0, self.num_repeats):
                cluster_list = self.cluster_dict[cluster_met](self.df.SMILES, n_clusters=num_clus)
                gkfs = uru.GroupKFoldShuffle(n_splits=self.num_splits, shuffle=True)
                for train, test in gkfs.split(X=np.stack(self.df.fp), groups=cluster_list):
                    res.append([num_clus, len(test)])
        return pd.DataFrame(
            res,
            columns=[
                "num_clusters",
                "num_test_mols",
            ]
        )

    def tanimoto_one_vs_all(
            self,
            fp_arr1,
            fp_arr2,
            top_n=5,
    ):
        arr = []
        for fp2 in fp_arr2:
            sim_list = BulkTanimotoSimilarity(fp2, fp_arr1)
            sim_array = np.array(sim_list)
            idx = np.argpartition(np.array(sim_array), -top_n)[-top_n:]
            best_n_tanimoto = sim_array[idx]
            arr.append(best_n_tanimoto)
        return np.array(arr).flatten()

    def sim_test_train(
            self,
            split_id_dict,
            met='tanimoto',
            top_n=5,
    ):
        res = []
        for cluster_met in self.cluster_mets:
            for repeat_num in range(self.num_repeats):
                for fold_i in range(self.num_splits):
                    df_train =self.df.iloc[split_id_dict[cluster_met][repeat_num][fold_i][0]]
                    df_test =self.df.iloc[split_id_dict[cluster_met][repeat_num][fold_i][1]]
                    if met == 'tanimoto':
                        sim_vals = self.tanimoto_one_vs_all(
                            fp_arr1=df_train.fp.tolist(),
                            fp_arr2=df_test.fp.tolist(),
                            top_n=top_n,
                        )
                    idx = np.arange(0, len(sim_vals)) + fold_i * len(sim_vals)
                    sim_df = pd.DataFrame({"sim": sim_vals, "fold": fold_i, "cluster_met": cluster_met, "idx": idx})
                    res.append(sim_df)
        df_sim = pd.concat(res)
        # print(df_sim)
        return df_sim


if __name__ == "__main__":
    from chembridger.path import to

    df = pd.read_csv(to("data/protac/finalPROTAC.csv"))

    p = Splitting(
        df=df,
        num_repeats=1,
        num_splits=5,
    )


__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"

import numpy as np
from rdkit.Chem import Crippen
from rdkit.ML.Descriptors import MoleculeDescriptors
from rdkit.Chem import Descriptors


class Physicochemical:

    def __init__(self):
        descriptors_interest = [
            'MaxEStateIndex', 'MinEStateIndex', 'MaxAbsEStateIndex',
            'MinAbsEStateIndex', 'qed', 'MolWt', 'HeavyAtomMolWt',
            'ExactMolWt', 'NumValenceElectrons', 'NumRadicalElectrons',
            'FpDensityMorgan1', 'FpDensityMorgan2',
            'FpDensityMorgan3', 'BalabanJ', 'BertzCT', 'Chi0', 'Chi0n',
            'Chi0v', 'Chi1', 'Chi1n', 'Chi1v', 'Chi2n', 'Chi2v', 'Chi3n',
            'Chi3v', 'Chi4n', 'Chi4v', 'HallKierAlpha', 'Kappa1',
            'Kappa2', 'Kappa3', 'LabuteASA', 'PEOE_VSA1', 'PEOE_VSA10',
            'PEOE_VSA11', 'PEOE_VSA12', 'PEOE_VSA13', 'PEOE_VSA14',
            'PEOE_VSA2', 'PEOE_VSA3', 'PEOE_VSA4', 'PEOE_VSA5', 'PEOE_VSA6',
            'PEOE_VSA7', 'PEOE_VSA8', 'PEOE_VSA9', 'SMR_VSA1', 'SMR_VSA10',
            'SMR_VSA2', 'SMR_VSA3', 'SMR_VSA4', 'SMR_VSA5', 'SMR_VSA6',
            'SMR_VSA7', 'SMR_VSA8', 'SMR_VSA9', 'SlogP_VSA1', 'SlogP_VSA10',
            'SlogP_VSA11', 'SlogP_VSA12', 'SlogP_VSA2', 'SlogP_VSA3',
            'SlogP_VSA4', 'SlogP_VSA5', 'SlogP_VSA6', 'SlogP_VSA7',
            'SlogP_VSA8', 'SlogP_VSA9', 'TPSA', 'EState_VSA1', 'EState_VSA10',
            'EState_VSA11', 'EState_VSA2', 'EState_VSA3', 'EState_VSA4',
            'EState_VSA5', 'EState_VSA6', 'EState_VSA7', 'EState_VSA8',
            'EState_VSA9', 'VSA_EState1', 'VSA_EState10', 'VSA_EState2',
            'VSA_EState3', 'VSA_EState4', 'VSA_EState5', 'VSA_EState6',
            'VSA_EState7', 'VSA_EState8', 'VSA_EState9', 'FractionCSP3',
            'HeavyAtomCount', 'NHOHCount', 'NOCount', 'NumAliphaticCarbocycles',
            'NumAliphaticHeterocycles', 'NumAliphaticRings',
            'NumAromaticCarbocycles', 'NumAromaticHeterocycles',
            'NumAromaticRings', 'NumHAcceptors', 'NumHDonors', 'NumHeteroatoms',
            'NumRotatableBonds', 'NumSaturatedCarbocycles',
            'NumSaturatedHeterocycles', 'NumSaturatedRings', 'RingCount',
            'MolLogP', 'MolMR', 'fr_Al_COO', 'fr_Al_OH', 'fr_Al_OH_noTert',
            'fr_ArN', 'fr_Ar_COO', 'fr_Ar_N', 'fr_Ar_NH', 'fr_Ar_OH', 'fr_COO',
            'fr_COO2', 'fr_C_O', 'fr_C_O_noCOO', 'fr_C_S', 'fr_HOCCN',
            'fr_Imine', 'fr_NH0', 'fr_NH1', 'fr_NH2', 'fr_N_O',
            'fr_Ndealkylation1', 'fr_Ndealkylation2', 'fr_Nhpyrrole',
            'fr_SH', 'fr_aldehyde', 'fr_alkyl_carbamate', 'fr_alkyl_halide',
            'fr_allylic_oxid', 'fr_amide', 'fr_amidine', 'fr_aniline',
            'fr_aryl_methyl', 'fr_azide', 'fr_azo', 'fr_barbitur',
            'fr_benzene', 'fr_benzodiazepine', 'fr_bicyclic', 'fr_diazo',
            'fr_dihydropyridine', 'fr_epoxide', 'fr_ester', 'fr_ether',
            'fr_furan', 'fr_guanido', 'fr_halogen', 'fr_hdrzine',
            'fr_hdrzone', 'fr_imidazole', 'fr_imide', 'fr_isocyan',
            'fr_isothiocyan', 'fr_ketone', 'fr_ketone_Topliss', 'fr_lactam',
            'fr_lactone', 'fr_methoxy', 'fr_morpholine', 'fr_nitrile',
            'fr_nitro', 'fr_nitro_arom', 'fr_nitro_arom_nonortho',
            'fr_nitroso', 'fr_oxazole', 'fr_oxime', 'fr_para_hydroxylation',
            'fr_phenol', 'fr_phenol_noOrthoHbond', 'fr_phos_acid',
            'fr_phos_ester', 'fr_piperdine', 'fr_piperzine', 'fr_priamide',
            'fr_prisulfonamd', 'fr_pyridine', 'fr_quatN', 'fr_sulfide',
            'fr_sulfonamd', 'fr_sulfone', 'fr_term_acetylene', 'fr_tetrazole',
            'fr_thiazole', 'fr_thiocyan', 'fr_thiophene', 'fr_unbrch_alkane',
            'fr_urea'
        ]
        self.allowed_descriptors = set(descriptors_interest)

    def real(self, mol):
        return mol.GetPropsAsDict()

    def dgbid(self, mol):
        return mol.GetProp('DATABASE_ID')

    def descriptors1(self, mol, is_dict=False):
        """
        ..  @description:
            -------------
            It is essentially same with the way of descriptors2().
            The difference is from the module of "MoleculeDescriptors".

        ..  @example:
            ---------
            1. mol_Bivalirudin = Chem.MolFromSmiles('CC[C@H](C)[C@H](NC(=O)[C@H](CCC(O)=O)NC(=O)[C@H](CCC(O)=O)NC(=O)[C@H](CC1=CC=CC=C1)NC(=O)[C@H](CC(O)=O)NC(=O)CNC(=O)[C@H](CC(N)=O)NC(=O)CNC(=O)CNC(=O)CNC(=O)CNC(=O)[C@@H]1CCCN1C(=O)[C@H](CCCNC(N)=N)NC(=O)[C@@H]1CCCN1C(=O)[C@H](N)CC1=CC=CC=C1)C(=O)N1CCC[C@H]1C(=O)N[C@@H](CCC(O)=O)C(=O)N[C@@H](CCC(O)=O)C(=O)N[C@@H](CC1=CC=C(O)C=C1)C(=O)N[C@@H](CC(C)C)C(O)=O')
            or
            2. mol_Bivalirudin = mols[0]

        :param mols:
        :return: a list (200)
        """
        names = [d[0] for d in Descriptors._descList]
        dpt = MoleculeDescriptors.MolecularDescriptorCalculator(names)
        descriptor = dpt.CalcDescriptors(mol)
        print(names)
        if is_dict:
            descriptors = dict()
            for i, v in enumerate(names):
                descriptors[v] = descriptor[i]
        else:
            descriptors = np.array(descriptor)
        return descriptors

    def descriptors2(self, mol, is_dict=False):
        """
        ..  @description
            ------------
            1. Descriptors._descList is a list consisting of elements each
                of which is a 2d array.
            2. It is essentially same with the way of descriptors1().

        :param mol: should be single molecule.
        :return: a dict or an array
        """
        descriptors_dict = {}
        _descList = []
        descriptors_list = []
        # print(len(Descriptors._descList))
        # print(len(self.allowedDescriptors))
        for descriptor, function in Descriptors._descList:
            if descriptor in self.allowed_descriptors:
                if is_dict:
                    descriptors_dict[descriptor] = function(mol)
                else:
                    descriptors_list.append(function(mol))
                    _descList.append((descriptor, function))
        if is_dict:
            descriptors = descriptors_dict
        else:
            descriptors = np.array(descriptors_list)
        return descriptors

    def crippen(self, mol, is_dict=False):
        LogP = Crippen.MolLogP(mol)
        molar_refractivity = Crippen.MolMR(mol)
        descriptor = [LogP, molar_refractivity]
        if is_dict:
            descriptors = {'LogP': LogP, 'MR': molar_refractivity}
        else:
            descriptors = np.array(descriptor)
        return descriptors


if __name__ == "__main__":
    from pypropel.path import to
    from pypropel.chem.SDF import SDF

    p1 = Physicochemical()
    mp = SDF(sdf_fpn=to('data/chem/approved.sdf'))
    mp.kind = 'all_dict'
    mols = mp.extract()

    # print(p1.dgbid(mols['DB09517']))

    # print(p1.real(mols['DB09517']))


    # print(p1.descriptors1(mols['DB09517'], is_dict=True))
    # print(p1.descriptors2(mols['DB01169'], is_dict=True))

    # print(p1.crippen(mols['DB09517'], is_dict=True))


__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "MIT"
__lab__ = "Cribbslab"

from datetime import datetime


class Console:

    def __init__(self, placeholder='logger: ', verbose=False):
        self._verbose = verbose
        self.placeholder = placeholder

    @property
    def verbose(self, ):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = value

    def print(self, content):
        if self._verbose:
            now = datetime.now()
            dt_format = now.strftime("%d/%m/%Y %H:%M:%S ")
            print(dt_format + self.placeholder + str(content))

    def check(self, content):
        if self._verbose:
            print(content)


__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"

import pandas as pd
from rdkit import DataStructs
from chembridger.util.Writer import Writer as pfwriter
from chembridger.util.Console import Console


class Similarity:

    def __init__(self, verbose=False):
        self.console = Console()
        self.verbose = verbose
        self.console.verbose = self.verbose

    def get(self, fps, type='Tanimoto', is_sv=False, sv_fpn='./'):
        """

        :param fps: a list of fingerprints - self.fingerprints(type='Morgan')
        :param type: Tanimoto, Dice, Cosine, Sokal, Russel,
                     Kulczynski, McConnaughey, and Tversky

        :return:
        """
        self.console.print('===>calculating similarity using {}'.format(type))
        num_fps = len(fps)
        if type == 'Tanimoto':
            method = DataStructs.TanimotoSimilarity
        elif type == 'Dice':
            method = DataStructs.DiceSimilarity
        elif type == 'Sokal':
            method = DataStructs.SokalSimilarity
        elif type == 'Russel':
            method = DataStructs.RusselSimilarity
        elif type == 'Kulczynski':
            method = DataStructs.KulczynskiSimilarity
        elif type == 'McConnaughey':
            method = DataStructs.McConnaugheySimilarity
        else:
            method = DataStructs.TverskySimilarity
        sims = []
        for i in range(num_fps):
            if i!= 0 and i % 1000 == 0:
                self.console.print('======>{} molecules processed'.format(i))
            for j in range(i+1, num_fps):
                sims.append([
                    fps[i][0],
                    fps[j][0],
                    DataStructs.FingerprintSimilarity(
                        fps[i][1], fps[j][1], metric=method
                    ),
                ])
        sims_ = pd.DataFrame(sims)
        if is_sv:
            pfwriter().generic(sims_, sv_fpn)
        return sims_


if __name__ == "__main__":
    from chembridger.path import to
    from chembridger.chem.SDF import SDF
    from chembridger.chem.Fingerprint import Fingerprint

    p1 = Similarity()
    p2 = Fingerprint()
    p3 = SDF(sdf_fpn=to('data/chem/approved.sdf'))
    # p3.kind_ = 'all_2d'
    # p3.kind_ = 'approved_2d'
    p3.kind = 'expt_2d'
    mols = p3.extract()
    print(len(mols))
    fps = p2.get(mols, is_to_vec=False, single=False)
    print(fps)

    print(p1.get(
        fps=fps,
        type='Tanimoto',
        is_sv=True,
        sv_fpn=to('data/chem/expt_2d.si'),
    ))



import numpy as np
import pandas as pd
# from statsmodels.stats.multitest import multipletests
# from joblib import Parallel, delayed

class KSScorer:
    def __init__(self, ref_matrix, gene_names=None, sample_names=None):
        if isinstance(ref_matrix, pd.DataFrame):
            self.ref_matrix = ref_matrix.to_numpy()
            self.gene_names = ref_matrix.index.to_numpy()
            self.sample_names = ref_matrix.columns.to_numpy()
        else:
            self.ref_matrix = ref_matrix
            self.gene_names = gene_names or np.array([f"gene{i+1}" for i in range(ref_matrix.shape[0])])
            self.sample_names = sample_names or np.array([f"sample{i+1}" for i in range(ref_matrix.shape[1])])
        self.ranked_lists = self._matrix_to_ranked_lists()

    def _matrix_to_ranked_lists(self):
        return [list(self.gene_names[np.argsort(-self.ref_matrix[:, i])]) for i in range(self.ref_matrix.shape[1])]

    def _ks_score(self, ref_list, query):
        len_ref = len(ref_list)
        query_rank = np.array([ref_list.index(gene) + 1 for gene in query if gene in ref_list])
        query_rank.sort()
        len_query = len(query_rank)
        if len_query == 0:
            return 0.0
        d = np.arange(1, len_query + 1) / len_query - query_rank / len_ref
        a = np.max(d)
        b = -np.min(d) + 1 / len_query
        return a if a > b else -b

    def _ks(self, ref_list, query_up, query_down):
        score_up = self._ks_score(ref_list, query_up)
        score_down = self._ks_score(ref_list, query_down)
        return score_up - score_down if score_up * score_down <= 0 else 0.0

    def compute(self, query_up, query_down, permute_num=1000, p_adj_method="fdr_bh", n_jobs=1):
        query_up = [gene for gene in query_up if gene in self.gene_names]
        query_down = [gene for gene in query_down if gene in self.gene_names]

        scores = Parallel(n_jobs=n_jobs)(
            delayed(self._ks)(ref_list, query_up, query_down) for ref_list in self.ranked_lists
        )
        scores = np.array(scores)

        permute_scores = np.zeros((self.ref_matrix.shape[1], permute_num))
        for n in range(permute_num):
            boot_up = np.random.choice(self.gene_names, size=len(query_up), replace=False)
            boot_down = np.random.choice(self.gene_names, size=len(query_down), replace=False)
            permute_scores[:, n] = Parallel(n_jobs=n_jobs)(
                delayed(self._ks)(ref_list, boot_up, boot_down) for ref_list in self.ranked_lists
            )

        permute_scores = np.nan_to_num(permute_scores)
        p_values = np.mean(np.abs(permute_scores) >= np.abs(scores[:, np.newaxis]), axis=1)
        p_adjusted = multipletests(p_values, method=p_adj_method)[1]

        result = pd.DataFrame({
            "Score": scores,
            "pValue": p_values,
            "pAdjValue": p_adjusted
        }, index=self.sample_names)

        return result



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

from chembridger.descriptor.Fingerprint import Fingerprint


def fingerprint(
        mols,
        met,
        **kwargs,
):
    return Fingerprint(
        mols=mols,
        met=met,
        **kwargs,
    ).get()


if __name__ == "__main__":
    from chembridger.path import to
    import pandas as pd

    df = pd.read_csv(to("data/protac/finalPROTAC.csv"))

    from smiles import to_mols
    df['mol'] = to_mols(
        smiles=df["SMILES"],
        verbose=True,
    )

    df['fp'] = fingerprint(
        mols=df["mol"].values.tolist(),
        met="Morgan",
    )
    print(df)


__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2024"
__license__ = "GPL-3.0"
__email__ = "jianfeng.sunmt@gmail.com"

import pandas as pd
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem.AtomPairs import Pairs
from rdkit.Chem.AtomPairs import Torsions
from rdkit.Avalon.pyAvalonTools import GetAvalonFP
from rdkit.Chem import rdFingerprintGenerator
from chembridger.util.Console import Console


class Fingerprint:

    def __init__(
            self,
            mols,
            met='Morgan',
            verbose=True,
            **kwargs,
    ):
        self.mols = mols
        self.met = met
        self.df = pd.DataFrame(self.mols, columns=['mol'])
        # print(self.df)
        # is_to_vec=False,

        self.console = Console()
        self.console.verbose = verbose

    def get(self, ):
        if self.met == 'Morgan':
            fpgen = rdFingerprintGenerator.GetMorganGenerator()
            # return self.df.mol.apply(fpgen.GetCountFingerprintAsNumPy)
            return self.df.mol.apply(fpgen.GetFingerprint)

    def get2(
            self,

    ):
        """

        Parameters
        ----------
        mols
        type
        self.dgb_mols_all or self.dgb_mols_approved
            1. MACCSKeys: 167bit,
            2. Morgan: changeable,
            3. Hashed_Torsion,
            4. Torsion,
            5. Atom_Pair,
            6. RDK,
            7. Avalon: 512 bit.
        is_to_vec
        single

        Returns
        -------

        """
        # print('===>calculating {} fingerprints'.format(type))
        if type == 'MACCSKeys':
            if single:
                fps = AllChem.GetMACCSKeysFingerprint(mols)
            else:
                fps = [[mol[0], AllChem.GetMACCSKeysFingerprint(mol[1])] for mol in mols]
        elif type == 'Morgan':
            if single:
                fps = AllChem.GetMorganFingerprintAsBitVect(mols, radius=2, nBits=1024)
            else:
                fps = [[mol[0], AllChem.GetMorganFingerprintAsBitVect(mol[1], radius=2, nBits=1024)] for mol in mols]
        elif type == 'Hashed_Torsion':
            if single:
                fps = rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect(mols)
            else:
                fps = [[mol[0], rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect(mol[1])] for mol in mols]
        elif type == 'Torsion':
            if single:
                fps = Torsions.GetTopologicalTorsionFingerprintAsIntVect(mols)
            else:
                fps = [[mol[0], Torsions.GetTopologicalTorsionFingerprintAsIntVect(mol[1])] for mol in mols]
        elif type == 'Atom_Pair':
            if single:
                fps = Pairs.GetAtomPairFingerprintAsBitVect(mols)
            else:
                fps = [[mol[0], Pairs.GetAtomPairFingerprintAsBitVect(mol[1])] for mol in mols]
        elif type == 'RDK':
            if single:
                fps = Chem.RDKFingerprint(mols, maxPath=5)
            else:
                fps = [[mol[0], Chem.RDKFingerprint(mol[1], maxPath=5)] for mol in mols]
        elif type == 'Avalon':
            if single:
                fps = GetAvalonFP(mols)
            else:
                fps = [[mol[0], GetAvalonFP(mol[1])] for mol in mols]
        else:
            if single:
                fps = Chem.RDKFingerprint(mols, maxPath=5)
            else:
                fps = [[mol[0], Chem.RDKFingerprint(mol[1], maxPath=5)] for mol in mols]
        if is_to_vec:
            if single:
                fps = fps.ToBitString()
            else:
                fps = [[fp[0], fp[1].ToBitString()] for i, fp in enumerate(fps)]
        return fps


if __name__ == "__main__":
    from chembridger.path import to
    from chembridger.molecule.SDF import SDF

    p = Fingerprint()
    mp = SDF(sdf_fpn=to('data/chem/approved.sdf'))
    mp.kind = 'all_dict' # approved
    mols = mp.extract()
    # print(mols)
    fps = p.get(
        mols['DB17308'],
        is_to_vec=True,
        type='Morgan',
        single=True
    )
    print(fps)

    # print(len(fps))

    print(p.get(mols, is_to_vec=True, single=False))